/* This file is part of Cloudy and is copyright (C)1978-2015 by Gary J. Ferland and
 * others.  For conditions of distribution and use see copyright notice in license.txt */
/*lines_setup convert level 1 and level 2 line parameters and pointers into internal form 
 * used by code, line data were read in by atmdat_readin */
#include "cddefines.h"
#include "prt.h"
#include "taulines.h"
#include "opacity.h"
#include "lines_service.h"

/* following used for generating array indices to level1 lines,
 * start process, use it, then end */
STATIC void initFindLevLine( void );
STATIC long ipFindLevLine( realnum , long , long );
STATIC void endFindLevLine( void );

void lines_setup(void)
{
	static bool lgFirst = true;
	bool lgSane;

	DEBUG_ENTRY( "lines_setup()" );

	/* this routine takes the line parameters in the wind block data and sorts
	 * them into what is needed for the actual line optical depth arrays */

	/** \todo	2	initialization already done at this point */

	/* this is the dummy line */
	TauLines[0].WLAng() = 0.;
	(*TauLines[0].Lo()).g() = 0.;
	(*TauLines[0].Hi()).g() = 0.;
	TauLines[0].Emis().gf() = 0.;
	TauLines[0].EnergyWN() = 0.;
	(*TauLines[0].Hi()).IonStg() = 0;
	(*TauLines[0].Hi()).nelem() = 0;
	/* this is an impossible value of iRedisFun() */
	TauLines[0].Emis().iRedisFun() = 0;
	TauLines[0].Emis().Aul() = 0.;

	/* the first valid line is [0] since zero is the dummy */
	if( TauLines[1].EnergyWN() <= 0. )
	{
		fprintf( ioQQQ, " PROBLEM Insane value for TauLines array.\n" );
		fprintf( ioQQQ, " Was block data LineData linked in??\n" );
		fprintf( ioQQQ, " Check that it compiled OK (it probably did not).\n" );
		TotalInsanity();
	}

	/* check that all lines have valid data */
	lgSane = true;
	for( long i=1; i <= nLevel1; i++ )
	{

		if( (*TauLines[i].Lo()).g() <= 0. )
		{
			fprintf( ioQQQ, "  routine lines_setup, insane lower stat wght\n" );
			fprintf( ioQQQ, " line index is %5ld\n", i );
			lgSane = false;
		}

		if( (*TauLines[i].Hi()).g() <= 0. )
		{
			fprintf( ioQQQ, "  routine lines_setup, insane upper stat wght\n" );
			fprintf( ioQQQ, " line index is %5ld\n", i );
			lgSane = false;
		}

		if( TauLines[i].EnergyWN() <= 0. )
		{
			fprintf( ioQQQ, "  routine lines_setup, insane energy WN\n" );
			fprintf( ioQQQ, " line index is %5ld\n", i );
			lgSane = false;
		}

		if( (*TauLines[i].Hi()).IonStg() <= 0 )
		{
			fprintf( ioQQQ, "  routine lines_setup, insane ioniz stage\n" );
			fprintf( ioQQQ, " line index is %5ld\n", i );
			lgSane = false;
		}

		if( (*TauLines[i].Hi()).nelem() <= 0 || (*TauLines[i].Hi()).nelem() > (int)LIMELM )
		{
			fprintf( ioQQQ, "  routine lines_setup, insane Nelem\n" );
			fprintf( ioQQQ, " line index is %5ld\n", i );
			lgSane = false;
		}

		if( (*TauLines[i].Hi()).IonStg() > (*TauLines[i].Hi()).nelem() )
		{
			fprintf( ioQQQ, "  routine lines_setup, insane IonStg>Nelem\n" );
			fprintf( ioQQQ, " line index is %5ld\n", i );
			lgSane = false;
		}

		if( TauLines[i].Emis().iRedisFun() == 0 )
		{
			fprintf( ioQQQ, "  routine lines_setup, insane line redis fcn\n" );
			fprintf( ioQQQ, " line index is %5ld\n", i );
			lgSane = false;
		}

		/* use energies for wavelengths in air if wl not forced with wl number on line */
		/* >>chng 03 oct 07, only make correction for index ref if
		 * if wl was not already set - this is an option to allow
		 * the printed wl to be specified in the level1.dat file */
		if( TauLines[i].WLAng() <= 0. )
		{
			/* make following an air wavelength */
			TauLines[i].WLAng() = 
				(realnum)(1.0e8/
							 TauLines[i].EnergyWN()/
							 RefIndex( TauLines[i].EnergyWN()));
		}
		{
			/*@-redef@*/
			enum{DEBUG_LOC=false};
			/*@+redef@*/
			if( DEBUG_LOC  )
			{
				fprintf( ioQQQ,"%s\n", chLineLbl(TauLines[i]).c_str() );
			}
		}
	}

	if( !lgSane )
	{
		fprintf( ioQQQ, " Insane value for line arrays encountered.\n" );
		fprintf( ioQQQ, " Was block data lines linked in??\n" );
		fprintf( ioQQQ, " Were errors intreoducted into the line array?\n" );
		ShowMe();
		cdEXIT(EXIT_FAILURE);
	}

	/* set up array to store hits for each line */
	initFindLevLine( );

	/* in following calls to ipFindLevLine the numbers are the integer wavelength
	 * used in the printout, the ion stage, and the atomic number */


	/*>>chng 03 oct 07, chng wl from 2360 to 2400, the value
	 * before cnhg of feii lines to stuc */
	ipTuv3 = ipFindLevLine( 2400 ,  2 , 26 );

	ipTr48 = ipFindLevLine( 6200 ,  2 , 26 );

	ipTFe16 = ipFindLevLine( 1080 ,  2 , 26 );

	ipTFe26 = ipFindLevLine( 1500 ,  2 , 26 );

	ipTFe34 = ipFindLevLine( 11500 ,  2 , 26 );

	ipTFe35 = ipFindLevLine( 2500 ,  2 , 26 );

	ipTFe46 = ipFindLevLine( 2300 ,  2 , 26 );

	ipTFe56 = ipFindLevLine( 8900 ,  2 , 26 );

	ipT191 = ipFindLevLine( 1786 , 2 , 26 );

	/* flush the line list, freeing the extra storage and checking that all
	 * lines have been claimed */
	endFindLevLine( );

	/** \todo	1	streamline all of this, using transition::Zero
	 * and then setting dangerously large negative numbers. */

	/* only do this one time, and only if number of atom_level2 lines is positive */
	if( lgFirst && nWindLine>0)
	{ 

		lgFirst = false;
		/* these are the massive set of op lines, with g-bar approx cs
		 * confirm that input data are valid */

		for( long i=0; i < nWindLine; i++ )
		{
			/* this information was read in in createdata */
			ASSERT( (*TauLine2[i].Hi()).nelem() > 0 );
			ASSERT( (*TauLine2[i].Hi()).nelem() <= (int)LIMELM );

			ASSERT( (*TauLine2[i].Hi()).IonStg() > 0 );
			ASSERT( (*TauLine2[i].Hi()).IonStg() <= (int)LIMELM );

			ASSERT( (*TauLine2[i].Lo()).g() >0. );

			ASSERT( (*TauLine2[i].Hi()).g() > 0. );

			/* check that energy is positive*/
			ASSERT( TauLine2[i].EnergyWN() > 0 );

			/* TauLine2[i].Emis().gf() this is gf if positive, A if negative */
			/* test whether a or gf entered, convert A to gf */
			if( TauLine2[i].Emis().gf() < 0. )
			{
				/* convert A (=-gf) into real gf */
				TauLine2[i].Emis().gf() *= (realnum)((-(*TauLine2[i].Hi()).g())/TRANS_PROB_CONST/POW2(TauLine2[i].EnergyWN()));
			}

			/*now put into standard format */
			TauLine2[i].WLAng() = 1.e8f/(TauLine2[i].EnergyWN() * RefIndex(TauLine2[i].EnergyWN()));
			(*TauLine2[i].Lo()).Pop() = 0.;
			(*TauLine2[i].Hi()).Pop() = 0.;
			TauLine2[i].Emis().iRedisFun() = ipPRD;

			/* these are line optical depth arrays
			 * inward optical depth */
			TauLine2[i].Emis().TauIn() = opac.taumin;
			TauLine2[i].Emis().TauCon() = opac.taumin;
			TauLine2[i].Emis().ColOvTot() = 0.;
			/* outward optical depth */
			TauLine2[i].Emis().TauTot() = 1e20f;
			/* escape probability */
			TauLine2[i].Emis().Pesc() = 1.;
			/* inward part of line */
			TauLine2[i].Emis().FracInwd() = 1.;
			/* destruction probability */
			TauLine2[i].Emis().Pdest() = 0.;
			TauLine2[i].Emis().Pelec_esc() = 0.;
			/* line pumping rate */
			TauLine2[i].Emis().pump() = 0.;
			/* population of lower level */
			(*TauLine2[i].Lo()).Pop() = 0.;
			/* population of upper level */
			(*TauLine2[i].Hi()).Pop() = 0.;
			/* population of lower level with correction for stim emission */
			TauLine2[i].Emis().PopOpc() = 0.;
			/* following two heat exchange excitation, deexcitation */
			TauLine2[i].Coll().cool() = 0.;
			TauLine2[i].Coll().heat() = 0.;
			/* intensity of line */
			TauLine2[i].Emis().xIntensity() = 0.;
			/* number of photons emitted in line */
			TauLine2[i].Emis().phots() = 0.;
			/* ots rate */
			TauLine2[i].Emis().ots() = 0.;
		}
	}

	for( size_t i=0; i < UTALines.size(); i++ )
	{
		/* this information was read in in createdata */
		ASSERT( (*UTALines[i].Hi()).nelem() > 0 );
		ASSERT( (*UTALines[i].Hi()).nelem() <= (int)LIMELM );

		ASSERT( (*UTALines[i].Hi()).IonStg() > 0 );
		ASSERT( (*UTALines[i].Hi()).IonStg() <= (int)LIMELM );

		ASSERT( (*UTALines[i].Lo()).g() > 0. );

		ASSERT( (*UTALines[i].Hi()).g() > 0. );

		/* check that energy is positive*/
		ASSERT( UTALines[i].EnergyWN() > 0 );

		(*UTALines[i].Lo()).Pop() = 0.;
		(*UTALines[i].Hi()).Pop() = 0.;
		UTALines[i].Emis().iRedisFun() = ipPRD;

		/* these are line optical depth arrays
		 * inward optical depth */
		UTALines[i].Emis().TauIn() = opac.taumin;
		UTALines[i].Emis().TauCon() = opac.taumin;
		UTALines[i].Emis().ColOvTot() = 0.;
		/* outward optical depth */
		UTALines[i].Emis().TauTot() = 1e20f;
		/* escape probability */
		UTALines[i].Emis().Pesc() = 1.;
		/* inward part of line */
		UTALines[i].Emis().FracInwd() = 1.;
		/* destruction probability */
		UTALines[i].Emis().Pdest() = 0.;
		UTALines[i].Emis().Pelec_esc() = 0.;
		/* line pumping rate */
		UTALines[i].Emis().pump() = 0.;
		/* population of lower level */
		(*UTALines[i].Lo()).Pop() = 0.;
		/* population of upper level */
		(*UTALines[i].Hi()).Pop() = 0.;
		/* population of lower level with correction for stim emission */
		UTALines[i].Emis().PopOpc() = 0.;
		/* following two heat exchange excitation, deexcitation */
		UTALines[i].Coll().cool() = 0.;
		/* heat is the net heat per pump and was set when data read in
		 * this is different from other lines with this structure 
		UTALines[i].Coll().heat() = 0.;*/
		/* intensity of line */
		UTALines[i].Emis().xIntensity() = 0.;
		/* number of photons emitted in line */
		UTALines[i].Emis().phots() = 0.;
		UTALines[i].Emis().ots() = 0.;
	}

	for( size_t i=0; i < HFLines.size(); i++ )
	{
		/* this information was read in in createdata */
		ASSERT( (*HFLines[i].Hi()).nelem() > 0 );
		ASSERT( (*HFLines[i].Hi()).nelem() <= (int)LIMELM );

		ASSERT( (*HFLines[i].Hi()).IonStg() > 0 );
		ASSERT( (*HFLines[i].Hi()).IonStg() <= (int)LIMELM );

		ASSERT( (*HFLines[i].Lo()).g() > 0. );

		ASSERT( (*HFLines[i].Hi()).g() > 0. );

		/* check that energy is positive*/
		ASSERT( HFLines[i].EnergyWN() > 0 );
		ASSERT( HFLines[i].Emis().Aul()>0 );
		ASSERT( HFLines[i].Emis().damp()>0 );

		/* HFLines[i].Emis->gf() this is gf if positive, A if negative */
		/* test whether a or gf entered, convert A to gf */
		if( HFLines[i].Emis().gf() < 0. )
		{
			/* convert A (=-gf) into real gf */
			HFLines[i].Emis().gf() *= (realnum)(-(*HFLines[i].Hi()).g()/TRANS_PROB_CONST/POW2(HFLines[i].EnergyWN()));
		}

		/*now put into standard format */
		HFLines[i].WLAng() = 1.e8f/HFLines[i].EnergyWN();
		(*HFLines[i].Lo()).Pop() = 0.;
		(*HFLines[i].Hi()).Pop() = 0.;
		/* change from partial to complete redistribution */
		HFLines[i].Emis().iRedisFun() = ipCRD;

		/* these are line optical depth arrays
		 * inward optical depth */
		HFLines[i].Emis().TauIn() = opac.taumin;
		HFLines[i].Emis().TauCon() = opac.taumin;
		HFLines[i].Emis().ColOvTot()=0;
		/* outward optical depth */
		HFLines[i].Emis().TauTot() = 1e20f;
		/* escape probability */
		HFLines[i].Emis().Pesc() = 1.;
		/* inward part of line */
		HFLines[i].Emis().FracInwd() = 1.;
		/* destruction probability */
		HFLines[i].Emis().Pdest() = 0.;
		HFLines[i].Emis().Pelec_esc() = 0.;
		/* line pumping rate */
		HFLines[i].Emis().pump() = 0.;
		/* population of lower level */
		(*HFLines[i].Lo()).Pop() = 0.;
		/* population of upper level */
		(*HFLines[i].Hi()).Pop() = 0.;
		/* population of lower level with correction for stim emission */
		HFLines[i].Emis().PopOpc() = 0.;
		/* following two heat exchange excitation, deexcitation */
		HFLines[i].Coll().cool() = 0.;
		HFLines[i].Coll().heat() = 0.;
		/* intensity of line */
		HFLines[i].Emis().xIntensity() = 0.;
		/* number of photons emitted in line */
		HFLines[i].Emis().phots() = 0.;
		HFLines[i].Emis().ots() = 0.;
	}
	return;
}

/* following used to save whether lines have been claimed by a pointer */
static int *lev2set;

/*generate pointer to level 1 line using wavelengtgh, ion, element */
STATIC long ipFindLevLine( 
		/* realnum ID wavelength, in angstroms */
		realnum wl , 
		/* state of ionization, 1 for neutral atom */
		long ion , 
		/* element number, 1 for H, 26 for Fe */
		long nelem)
{
	long i;/* use for counter in for loop */

	DEBUG_ENTRY( "ipFindLevLine()" );

	ASSERT( wl > 0. );
	ASSERT( ion > 0 );
	ASSERT( ion <= LIMELM );
	ASSERT( nelem > 0 );
	ASSERT( nelem <= LIMELM );

	/* look for the line */
	for( i=1; i<= nLevel1; ++i )
	{
		if( (*TauLines[i].Hi()).nelem() == (int)nelem &&
		    (*TauLines[i].Hi()).IonStg() == (int)ion &&
			// wl hit within relative accuracy of 0.05A in 1000A
		    fabs(TauLines[i].WLAng() - wl)/max(1000.,wl) < 0.05/1000. )
		{
			/* remember that we have hit this line */
			lev2set[i] = true;
			/* and return pointer to the label*/
			return i;
		}
	}
	fprintf(ioQQQ,
		" ipFindLevLine could not find a line with following properties:\n"
		" wavelength=%f\n"
		" ion stage =%li\n"
		" atomic num=%li\n",
		wl , ion, nelem );
	return -1;
}

STATIC void initFindLevLine( void )
{
	DEBUG_ENTRY( "initFindLevLine()" );

	/* generate the array of ints to store true and false */
	lev2set = (int*)MALLOC( (size_t)(nLevel1+1)*sizeof(int) );

	/* set them all false, saying that they have not been claimed by
	 * one of the line pointers */
	for( long i=1; i<=nLevel1; ++i )
		lev2set[i] = false;
	return;
}

STATIC void endFindLevLine( void )
{
	bool lgAbort_loc=false;

	DEBUG_ENTRY( "endFindLevLine()" );

	/* set them all false, saying that they have not been claimed by
	 * one of the line pointers */
	for( long i=1; i<=nLevel1; ++i )
	{
		if( !lev2set[i] )
		{
			fprintf(ioQQQ,"PROBLEM endFindLevLine warning; line %li not claimed\n",i);
			fprintf(ioQQQ,
				" line had the following properties:\n"
				" wavelength=%f\n"
				" ion stage =%i\n"
				" atomic num=%i\n",
				TauLines[i].WLAng() ,
					  (*TauLines[i].Hi()).IonStg() ,
					  (*TauLines[i].Hi()).nelem() );  
			lgAbort_loc = true;
		}
	}

	/* generate the array of ints to store true and false */
	free(lev2set);

	if( lgAbort_loc )
	{
		fprintf(ioQQQ," problems found entering the data.  I live in lines_setup.c\n");
		cdEXIT(EXIT_FAILURE);
	}
	return;
}
